--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- Mock Turtle
-- https://gitlab.cern.ch/coht/mockturtle
--------------------------------------------------------------------------------
--
-- unit name:   spec_mt_demo
--
-- description: Top entity for Mock Turtle SPEC demo.
--
-- This demo instantiates a Mock Turtle CPU with one core, connected to
-- a WB GPIO peripheral. The buttons and LEDS of the SPEC board are wired to
-- pins of the WB GPIO peripheral.
--
--------------------------------------------------------------------------------
-- Copyright (c) 2018 CERN
--------------------------------------------------------------------------------
-- GNU LESSER GENERAL PUBLIC LICENSE
--------------------------------------------------------------------------------
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.gencores_pkg.all;
use work.wishbone_pkg.all;
use work.gn4124_core_pkg.all;
use work.mock_turtle_pkg.all;
use work.mt_mqueue_pkg.all;
use work.synthesis_descriptor.all;

library unisim;
use unisim.vcomponents.all;

entity spec_mt_demo is
  port (
    -- resets and clocks
    rst_n_a_i           : in std_logic;
    clk_125m_pllref_p_i : in std_logic;
    clk_125m_pllref_n_i : in std_logic;

    -- Gennum PCIe
    -- gpio[0] -> gn4124 gpio8
    -- gpio[1] -> gn4124 gpio9
    gpio       : inout std_logic_vector(1 downto 0);
    l_rst_n    : in    std_logic;
    p2l_rdy    : out   std_logic;
    p2l_clkn   : in    std_logic;
    p2l_clkp   : in    std_logic;
    p2l_data   : in    std_logic_vector(15 downto 0);
    p2l_dframe : in    std_logic;
    p2l_valid  : in    std_logic;
    p_wr_req   : in    std_logic_vector(1 downto 0);
    p_wr_rdy   : out   std_logic_vector(1 downto 0);
    rx_error   : out   std_logic;
    l2p_data   : out   std_logic_vector(15 downto 0);
    l2p_dframe : out   std_logic;
    l2p_valid  : out   std_logic;
    l2p_clkn   : out   std_logic;
    l2p_clkp   : out   std_logic;
    l2p_edb    : out   std_logic;
    l2p_rdy    : in    std_logic;
    l_wr_rdy   : in    std_logic_vector(1 downto 0);
    p_rd_d_rdy : in    std_logic_vector(1 downto 0);
    tx_error   : in    std_logic;
    vc_rdy     : in    std_logic_vector(1 downto 0);

    -- SPEC buttons and LEDs
    button1_i : in  std_logic := '1';
    button2_i : in  std_logic := '1';
    leds_n_o  : out std_logic_vector(3 downto 0));

end entity spec_mt_demo;

architecture arch of spec_mt_demo is

  -- Host Message Queue configuration
  -- Each CPU has three queues assigned:
  -- 1) a small incoming queue (8 entries x 32 words) for sending host
  -- to CPU commands
  -- 2) a large outgoing queue (128 entries x 16 words) for optional event
  -- logging (asynchronous)
  -- 3) a small outgoing queue (8 entries x 128 words) for receiving
  -- command replies (CPU to host)
  constant c_hmq_config : t_mt_mqueue_config := (
    out_slot_count  => 2,
    out_slot_config => (
      0             => (width => 128, entries => 8),  -- control CPU 0 (to host)
      1             => (width => 16, entries => 128),  -- log CPU 0
      others        => (0, 0)),
    in_slot_count   => 1,
    in_slot_config  => (
      0             => (width => 32, entries => 8),  -- control CPU 0 (from host)
      others        => (0, 0)));

  constant c_mt_config : t_mock_turtle_config := (
    app_id          => x"d330d330",
    cpu_count       => 1,
    cpu_memsizes    => (32768, 0, 0, 0, 0, 0, 0, 0),
    hmq_config      => c_hmq_config,
    rmq_config      => c_dummy_mt_mqueue_config,
    shared_mem_size => 8192,
    tpu_enable      => FALSE,
    tpu_channels    => 1,
    tpu_buffer_size => 256);

  constant c_NUM_WB_MASTERS : integer := 2;
  constant c_NUM_WB_SLAVES  : integer := 1;

  constant c_MASTER_GENNUM : integer := 0;

  constant c_SLAVE_VIC      : integer := 0;
  constant c_SLAVE_MT       : integer := 1;
  constant c_DESC_SYNTHESIS : integer := 2;
  constant c_DESC_REPO_URL  : integer := 3;

  constant c_INTERCONNECT_LAYOUT : t_sdb_record_array(c_NUM_WB_MASTERS + 1 downto 0) := (
    c_SLAVE_VIC      => f_sdb_embed_device(c_xwb_vic_sdb, x"00002000"),
    c_SLAVE_MT       => f_sdb_embed_device(c_mock_turtle_sdb, x"00020000"),
    c_DESC_SYNTHESIS => f_sdb_embed_synthesis(c_sdb_synthesis_info),
    c_DESC_REPO_URL  => f_sdb_embed_repo_url(c_sdb_repo_url));

  constant c_SDB_ADDRESS : t_wishbone_address := x"00000000";

  constant c_VIC_VECTOR_TABLE : t_wishbone_address_array(0 to 1) := (
    0 => x"00020000",   -- MT Mqueue interrupt
    1 => x"00020001");  -- MT Debug Msg interrupt

  signal cnx_master_out : t_wishbone_master_out_array(c_NUM_WB_MASTERS-1 downto 0);
  signal cnx_master_in  : t_wishbone_master_in_array(c_NUM_WB_MASTERS-1 downto 0);

  signal cnx_slave_out : t_wishbone_slave_out_array(c_NUM_WB_SLAVES-1 downto 0);
  signal cnx_slave_in  : t_wishbone_slave_in_array(c_NUM_WB_SLAVES-1 downto 0);

  signal gn_wb_adr : std_logic_vector(31 downto 0);

  signal clk_125m_pllref : std_logic;
  signal clk_sys         : std_logic;

  signal rst_n_sys     : std_logic;
  signal local_reset_n : std_logic;

  signal pllout_clk_sys       : std_logic;
  signal pllout_clk_fb_pllref : std_logic;

  signal dp_wb_out : t_wishbone_master_out;
  signal dp_wb_in  : t_wishbone_master_in;

  signal cpu0_gpio_oen : std_logic_vector(5 downto 0);
  signal cpu0_gpio_out : std_logic_vector(5 downto 0);
  signal cpu0_gpio_in  : std_logic_vector(5 downto 0);

  signal mt_host_msg_irq  : std_logic;
  signal mt_debug_msg_irq : std_logic;
  signal vic_master_irq   : std_logic;

  signal powerup_reset_cnt : unsigned(7 downto 0) := "00000000";
  signal powerup_rst_n     : std_logic            := '0';
  signal sys_locked        : std_logic;

  -- only necessary to prevent errors due to unconnected
  -- inout port of GPIO peripherals
  signal dummy : std_logic_vector(5 downto 0);

begin  -- architecture arch

  U_Buf_CLK_PLL : IBUFGDS
    generic map (
      DIFF_TERM    => TRUE,
      IBUF_LOW_PWR => TRUE)
    port map (
      O  => clk_125m_pllref,
      I  => clk_125m_pllref_p_i,
      IB => clk_125m_pllref_n_i);

  cmp_sys_clk_pll : PLL_BASE
    generic map (
      BANDWIDTH          => "OPTIMIZED",
      CLK_FEEDBACK       => "CLKFBOUT",
      COMPENSATION       => "INTERNAL",
      DIVCLK_DIVIDE      => 1,
      CLKFBOUT_MULT      => 8,
      CLKFBOUT_PHASE     => 0.000,
      CLKOUT0_DIVIDE     => 16,  -- 62.5 MHz
      CLKOUT0_PHASE      => 0.000,
      CLKOUT0_DUTY_CYCLE => 0.500,
      CLKIN_PERIOD       => 8.0,
      REF_JITTER         => 0.016)
    port map (
      CLKFBOUT => pllout_clk_fb_pllref,
      CLKOUT0  => pllout_clk_sys,
      LOCKED   => sys_locked,
      RST      => '0',
      CLKFBIN  => pllout_clk_fb_pllref,
      CLKIN    => clk_125m_pllref);

  U_Sync_Reset : gc_sync_ffs
    port map (
      clk_i    => clk_sys,
      rst_n_i  => '1',
      data_i   => powerup_rst_n,
      synced_o => local_reset_n);

  cmp_clk_sys_buf : BUFG
    port map (
      O => clk_sys,
      I => pllout_clk_sys);

  U_GN4124_Core : gn4124_core
    port map (
      rst_n_a_i     => l_rst_n,
      status_o      => open,
      p2l_clk_p_i   => p2l_clkp,
      p2l_clk_n_i   => p2l_clkn,
      p2l_data_i    => p2l_data,
      p2l_dframe_i  => p2l_dframe,
      p2l_valid_i   => p2l_valid,
      p2l_rdy_o     => p2l_rdy,
      p_wr_req_i    => p_wr_req,
      p_wr_rdy_o    => p_wr_rdy,
      rx_error_o    => rx_error,
      vc_rdy_i      => vc_rdy,
      l2p_clk_p_o   => l2p_clkp,
      l2p_clk_n_o   => l2p_clkn,
      l2p_data_o    => l2p_data,
      l2p_dframe_o  => l2p_dframe,
      l2p_valid_o   => l2p_valid,
      l2p_edb_o     => l2p_edb,
      l2p_rdy_i     => l2p_rdy,
      l_wr_rdy_i    => l_wr_rdy,
      p_rd_d_rdy_i  => p_rd_d_rdy,
      tx_error_i    => tx_error,
      dma_irq_o     => open,
      irq_p_i       => '0',
      irq_p_o       => open,
      dma_reg_clk_i => clk_sys,
      csr_clk_i     => clk_sys,
      csr_adr_o     => gn_wb_adr,
      csr_dat_o     => cnx_slave_in(c_MASTER_GENNUM).dat,
      csr_sel_o     => cnx_slave_in(c_MASTER_GENNUM).sel,
      csr_stb_o     => cnx_slave_in(c_MASTER_GENNUM).stb,
      csr_we_o      => cnx_slave_in(c_MASTER_GENNUM).we,
      csr_cyc_o     => cnx_slave_in(c_MASTER_GENNUM).cyc,
      csr_dat_i     => cnx_slave_out(c_MASTER_GENNUM).dat,
      csr_ack_i     => cnx_slave_out(c_MASTER_GENNUM).ack,
      csr_stall_i   => cnx_slave_out(c_MASTER_GENNUM).stall,
      csr_err_i     => cnx_slave_out(c_MASTER_GENNUM).err,
      csr_rty_i     => cnx_slave_out(c_MASTER_GENNUM).rty,
      csr_int_i     => '0',
      dma_clk_i     => clk_sys,
      dma_ack_i     => '1',
      dma_stall_i   => '0',
      dma_err_i     => '0',
      dma_rty_i     => '0',
      dma_dat_i     => (others => '0'),
      dma_int_i     => '0',
      dma_reg_adr_i => (others => '0'),
      dma_reg_dat_i => (others => '0'),
      dma_reg_sel_i => (others => '0'),
      dma_reg_stb_i => '0',
      dma_reg_cyc_i => '0',
      dma_reg_we_i  => '0');

  cnx_slave_in(c_MASTER_GENNUM).adr <= gn_wb_adr(29 downto 0) & "00";

  U_Intercon : xwb_sdb_crossbar
    generic map (
      g_num_masters => c_NUM_WB_SLAVES,
      g_num_slaves  => c_NUM_WB_MASTERS,
      g_registered  => TRUE,
      g_wraparound  => TRUE,
      g_layout      => c_INTERCONNECT_LAYOUT,
      g_sdb_addr    => c_SDB_ADDRESS)
    port map (
      clk_sys_i => clk_sys,
      rst_n_i   => local_reset_n,
      slave_i   => cnx_slave_in,
      slave_o   => cnx_slave_out,
      master_i  => cnx_master_in,
      master_o  => cnx_master_out);

  U_VIC : xwb_vic
    generic map (
      g_interface_mode      => PIPELINED,
      g_address_granularity => BYTE,
      g_num_interrupts      => 2,
      g_init_vectors        => c_VIC_VECTOR_TABLE)
    port map (
      clk_sys_i    => clk_sys,
      rst_n_i      => local_reset_n,
      slave_i      => cnx_master_out(c_SLAVE_VIC),
      slave_o      => cnx_master_in(c_SLAVE_VIC),
      irqs_i(0)    => mt_host_msg_irq,
      irqs_i(1)    => mt_debug_msg_irq,
      irq_master_o => vic_master_irq);

  gpio(0) <= vic_master_irq;

  U_Mock_Turtle : mock_turtle_core
    generic map (
      g_config            => c_mt_config,
      g_with_rmq          => TRUE,
      g_with_white_rabbit => FALSE)
    port map (
      clk_i           => clk_sys,
      rst_n_i         => local_reset_n,
      dp_master_o(0)  => dp_wb_out,
      dp_master_i(0)  => dp_wb_in,
      host_slave_i    => cnx_master_out(c_SLAVE_MT),
      host_slave_o    => cnx_master_in(c_SLAVE_MT),
      host_irq_o      => mt_host_msg_irq,
      debug_msg_irq_o => mt_debug_msg_irq);

  U_GPIO_CPU0 : xwb_gpio_port
    generic map (
      g_interface_mode         => PIPELINED,
      g_address_granularity    => BYTE,
      g_num_pins               => 6,
      g_with_builtin_tristates => FALSE)
    port map (
      clk_sys_i  => clk_sys,
      rst_n_i    => rst_n_sys,
      slave_i    => dp_wb_out,
      slave_o    => dp_wb_in,
      gpio_b     => dummy,
      gpio_out_o => cpu0_gpio_out,
      gpio_in_i  => cpu0_gpio_in,
      gpio_oen_o => cpu0_gpio_oen);

  leds_n_o(0) <= not cpu0_gpio_out(0);
  leds_n_o(1) <= not cpu0_gpio_out(1);
  leds_n_o(2) <= not cpu0_gpio_out(2);
  leds_n_o(3) <= not cpu0_gpio_out(3);

  cpu0_gpio_in(4) <= not button1_i;
  cpu0_gpio_in(5) <= not button2_i;

  cpu0_gpio_in(3 downto 0) <= cpu0_gpio_out(3 downto 0);

end architecture arch;
