/**
 * @defgroup library-message-queue Message Queue
 * @{
 * @ingroup library
 * @brief Message Queues definitions and functions
 * @copyright (C) 2013-2016 CERN (www.cern.ch)
 * @author Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 * @author Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __TRTL_RT_MQUEUE_H
#define __TRTL_RT_MQUEUE_H

/**
 * Local MQ polling 32bit register. The Mock Turtle update this register whenever
 * there is an unread message in a MQ.
 * The bits [0, 15] are for the HMQ, bits [16,31] are for the RMQ.
 */
#define REG_LR_POLL    0x100000

/**
 * Host MQ base address
 */
#define HMQ_BASE           0x40010000
/**
 * Remote MQ base address
 */
#define RMQ_BASE           0x40020000

/**
 * MQ Global Control Register.
 * Offset from the MQ core base address.
 */
#define MQ_GCR       (0x0)

/**
 * It computes the input MQ address for a given slot.
 */
#define MQ_IN(slot)  (0x4000 + (slot) * 0x400)

/**
 * It computes the output MQ address for a given slot.
 */
#define MQ_OUT(slot) (0x8000 + (slot) * 0x400)

/**
 * MQ command claim. It will claim a MQ slot for exclusive use.
 */
#define MQ_CMD_CLAIM (1 << 24)

/**
 * MQ command purge.
 * It discards all messages in a slot
 */
#define MQ_CMD_PURGE (1 << 25)

/**
 * MQ command ready. It marks the message as ready to send.
 * The command need the message size argument.
 * The MQ slot content will be enqueued and sent by Mock Turtle
 */
#define MQ_CMD_READY (1 << 26)

/**
 * MQ command discard. It discards all messages in the slot.
 */
#define MQ_CMD_DISCARD (1 << 27)

/**
 * MQ Slot Command Register.
 * Offset from the MQ core base address.
 */
#define MQ_SLOT_COMMAND 0

/**
 * MQ Slot Status Register.
 * Offset from the MQ core base address.
 */
#define MQ_SLOT_STATUS 4

/**
 * MQ Slot Data Start Register.
 * Offset from the MQ core base address.
 */
#define MQ_SLOT_DATA_START 8

/**
 * Pointer to input buffer
 */
#define MQ_BUFFER_IN(_type, _slot) (trtl_mq_base_address(_type) + \
				    MQ_IN(_slot) +                \
				    MQ_SLOT_DATA_START)
/**
 * Pointer to output buffer
 */
#define MQ_BUFFER_OUT(_type, _slot) (trtl_mq_base_address(_type) + \
				    MQ_OUT(_slot) +                \
				    MQ_SLOT_DATA_START)


/**
 * List of Message Queue types
 */
enum trtl_mq_type {
	TRTL_HMQ = 0, /**< Host Message Queue - Host-Firmware communication */
	TRTL_RMQ, /**< Remote Message Queue - Network-Firmware communication */
};

/**
 * It gets the Message Queue base address
 * @param[in] type MQ type
 */
static inline void *trtl_mq_base_address(enum trtl_mq_type type)
{
	return (void *)(0x40010000 + ((type) * 0x10000));
}


/**
 * Remote Message Queue destination address
 */
struct rmq_message_addr {
	uint32_t target_ip; /**< destination IP addres*/
	uint32_t target_port; /**< destination port */
	uint32_t target_offset; /**< FIXME */
};


/**
 * It writes on a Message Queue register
 * @param[in] type MQ type to use
 * @param[val] value to write
 * @param[reg] reg register offset
 */
static inline void mq_writel(enum trtl_mq_type type, uint32_t val, uint32_t reg)
{
	*(volatile uint32_t *)(trtl_mq_base_address(type) + reg) = val;
}


/**
 * @copydoc MQ_CMD_CLAIM
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void mq_claim(enum trtl_mq_type type, int slot)
{
	mq_writel(type, MQ_CMD_CLAIM, MQ_OUT(slot) + MQ_SLOT_COMMAND);
}

/**
 * @copydoc MQ_CMD_PURGE
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void mq_purge(enum trtl_mq_type type, int slot)
{
	mq_writel(type, MQ_CMD_PURGE, MQ_OUT(slot) + MQ_SLOT_COMMAND);
}


/**
 * @copydoc MQ_CMD_READY
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 * @param[in] count number of 32bit words to send
 */
static inline void mq_send(enum trtl_mq_type type, int slot, int count)
{
	mq_writel(type, MQ_CMD_READY | count, MQ_OUT(slot) + MQ_SLOT_COMMAND);
}


/**
 * @copydoc MQ_CMD_DISCARD
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void mq_discard(enum trtl_mq_type type, int slot)
{
	mq_writel(type, MQ_CMD_DISCARD, MQ_IN(slot));
}


/**
 * It gets the output slot data field pointer
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static void *mq_map_out_buffer(enum trtl_mq_type type, int slot)
{
	return (void *) (trtl_mq_base_address(type) +
			 MQ_OUT (slot) +
			 MQ_SLOT_DATA_START);
}


/**
 * It gets the input slot data field pointer
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static void *mq_map_in_buffer(enum trtl_mq_type type, int slot)
{
	return (void *) (trtl_mq_base_address(type) +
			 MQ_IN (slot) +
			 MQ_SLOT_DATA_START);
}


/**
 * It gets the current status of the Message Queues (both Host and Remote)
 */
static inline uint32_t mq_poll(void)
{
	return *(volatile uint32_t *) ( REG_LR_POLL );
}


/**
 * It gets the current status of the Remote Message Queues
 */
static inline uint32_t rmq_poll(int slot)
{
	return *(volatile uint32_t *) ( REG_LR_POLL ) & ( 1<< (16 + slot ));
}

#endif
/**@}*/
